﻿using System;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Xml.Linq;

namespace PACX
{
    internal class PACXConnection
    {
        #region Private
        private bool connected;
        private DateTime last_received;
        private DateTime last_ping;
        private Socket socket;
        private SocketError error = new SocketError();
        private byte[] buffer = new byte[1024];
        private const string end_sequence = "\r\n\r\n\\;";
        private string pending_input = "";

        private void handle_input(string input)
        {
            last_received = DateTime.UtcNow;
            pending_input += input;
        }

        private void receiveCallback(IAsyncResult res)
        {
            try
            {
                int recv = socket.EndReceive(res);
                string inp = Encoding.Unicode.GetString(buffer, 0, recv);
                handle_input(inp);

                socket.BeginReceive(buffer, 0, buffer.Length, SocketFlags.None, out error, receiveCallback, null);
            }
            catch
            {
                OnPACXConnectionEvent?.Invoke(this, PACXConnectionEventType.Disconnected);
                Disconnect();
            }
        }
        #endregion

        #region Public
        internal delegate void PACXDataReceived(PACXConnection sender, string data);
        internal event PACXDataReceived OnPACXDataReceived;

        internal enum PACXConnectionEventType
        {
            Connected,
            ConnectionFailed,
            Disconnected
        }
        internal delegate void PACXConnectionEvent(PACXConnection sender, PACXConnectionEventType eventType);
        internal event PACXConnectionEvent OnPACXConnectionEvent;

        internal PACXConnection()
        {
            last_ping = DateTime.MinValue;
            last_received = DateTime.MinValue;
            connected = false;
        }

        internal void Connect()
        {
            string ip = "127.0.0.1";
            int port = 9000;

            try
            {
                string dir = System.IO.Path.Combine(System.IO.Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.LocalApplicationData), "PACX"), "Settings.xml");
                XElement xel = XElement.Load(dir);
                foreach (XElement setting in xel.Descendants("Settings")?.Elements())
                {
                    switch (setting.Name.ToString().ToLower())
                    {
                        case "networkip":
                            {
                                ip = setting.Value;
                            }
                            break;
                        case "networkport":
                            {
                                int.TryParse(setting.Value, out port);
                            }
                            break;
                    }
                }
            }
            catch { }

            Connect(ip, port);
        }

        internal void Connect(string addr, int port)
        {
            try
            {
                Disconnect();
                socket = new Socket(SocketType.Stream, ProtocolType.Tcp);
                socket.BeginConnect(new IPEndPoint(IPAddress.Parse(addr), port), (IAsyncResult res) =>
                {
                    try
                    {
                        socket.EndConnect(res);
                        if (socket.Connected)
                        {
                            last_received = DateTime.UtcNow;
                            last_ping = DateTime.UtcNow;
                            connected = true;
                            OnPACXConnectionEvent?.Invoke(this, PACXConnectionEventType.Connected);
                            socket.BeginReceive(buffer, 0, buffer.Length, SocketFlags.None, out error, receiveCallback, null);
                        }
                        else
                        {
                            OnPACXConnectionEvent?.Invoke(this, PACXConnectionEventType.ConnectionFailed);
                        }
                    }
                    catch
                    {
                        OnPACXConnectionEvent?.Invoke(this, PACXConnectionEventType.ConnectionFailed);
                    }
                }, null);
            }
            catch
            {
                OnPACXConnectionEvent?.Invoke(this, PACXConnectionEventType.ConnectionFailed);
            }
        }

        internal void Send(string message)
        {
            try
            {
                if (socket != null && socket.Connected)
                {
                    byte[] buf = Encoding.Unicode.GetBytes(message.Replace(end_sequence, "\r\n\r\n;") + end_sequence);
                    socket.BeginSend(buf, 0, buf.Length, SocketFlags.None, (IAsyncResult res) =>
                    {
                        try
                        {
                            socket.EndSend(res);
                        }
                        catch
                        {
                            OnPACXConnectionEvent?.Invoke(this, PACXConnectionEventType.Disconnected);
                            Disconnect();
                        }
                    }, null);
                }
            }
            catch
            {
                OnPACXConnectionEvent?.Invoke(this, PACXConnectionEventType.Disconnected);
                Disconnect();
            }
        }

        internal void Update()
        {
            if (!connected)
            {
                return;
            }
            if (DateTime.UtcNow - last_received > TimeSpan.FromSeconds(10.0))
            {
                OnPACXConnectionEvent?.Invoke(this, PACXConnectionEventType.Disconnected);
                Disconnect();
            }
            else
            {
                if (DateTime.UtcNow - last_ping > TimeSpan.FromSeconds(5.0))
                {
                    Send("ping");
                    last_ping = DateTime.UtcNow;
                }
            }
            string[] vals = pending_input.Split(new string[] { end_sequence }, StringSplitOptions.RemoveEmptyEntries);
            bool use_last = pending_input.EndsWith(end_sequence);
            for (int i = 0; i < vals.Length; i++)
            {
                if ((use_last == true || i < vals.Length - 1) && vals[i] != "pong")
                {
                    OnPACXDataReceived?.Invoke(this, vals[i]);
                }
            }
            if (use_last == false)
            {
                pending_input = vals[vals.Length - 1];
            }
            else
            {
                pending_input = "";
            }
        }

        internal bool IsConnected()
        {
            return connected;
        }

        internal void Disconnect()
        {
            try
            {
                if (socket != null)
                {
                    socket.Close();
                    socket = null;
                }
            }
            catch { }
            finally
            {
                connected = false;
            }
        }

        ~PACXConnection()
        {
            Disconnect();
            socket = null;
        }
        #endregion
    }
}
